@@ -1,4 +1,3 @@
-
 from typing import Dict, Any, List, Callable, get_type_hints, Optional, Union
 import inspect
 import ast
@@ -109,23 +108,86 @@ def _get_subparser_items(obj, subparser_name: str) -> Dict[str, Any]:
 class CommandWrapper:
     """Base wrapper for command functionality."""
     
-    def __init__(self, obj, name="", parent=None, allow_eval=False, include_subparser_name=False, validate_options = False):
+    def __init__(self, obj, name="", parent=None, allow_eval=False, include_subparser_name=False, 
+                 validate_options=False, flat=False, default_help=True):
         self._obj = obj
         self._name = name
         self._parent = parent
         self._allow_eval = allow_eval
         self._include_subparser_name = include_subparser_name
         self._validate_options = validate_options
+        self._flat = flat
         self._tree_cache = None
         self._commands = self._get_commands_for_obj(obj)
+        self._default_help = default_help
+        # Store subparser objects for flat mode
+        self._subparser_objects = self._get_subparser_objects(obj) if flat else {}
 
     def _get_commands_for_obj(self, obj: Any) -> dict:
         """Get all command methods for an object."""
         commands = {}
         for name, member in inspect.getmembers(obj):
             if hasattr(member, '_is_command'):
+                if member._when and not member._when(obj):
+                    continue
                 commands[name] = member
+                
+        # If flat mode is enabled, include commands from subparsers
+        if hasattr(self, '_flat') and self._flat:
+            subparser_commands = self._get_flat_commands(obj)
+            # Add subparser commands, but don't override existing commands
+            for name, cmd in subparser_commands.items():
+                if name not in commands:
+                    commands[name] = cmd
+                    
         return commands
+    
+    def _get_flat_commands(self, obj: Any) -> dict:
+        """Get all commands from subparsers when in flat mode."""
+        flat_commands = {}
+        
+        for subparser_name in getattr(obj.__class__, '_subparsers', []):
+            try:
+                subparser_items = _get_subparser_items(obj, subparser_name)
+                
+                for subparser_key, subparser_obj in subparser_items.items():
+                    # Get all commands from this subparser
+                    for name, member in inspect.getmembers(subparser_obj):
+                        if hasattr(member, '_is_command'):
+                            if member._when and not member._when(subparser_obj):
+                                continue
+                            # Store the command and its parent object for execution
+                            flat_commands[name] = member
+                            # Keep track of subparser objects for command execution
+                            if not hasattr(self, '_subparser_objects'):
+                                self._subparser_objects = {}
+                            if name not in self._subparser_objects:
+                                self._subparser_objects[name] = subparser_obj
+            except ValueError:
+                pass
+                
+        return flat_commands
+    
+    def _get_subparser_objects(self, obj: Any) -> dict:
+        """Get all subparser objects for flat mode."""
+        subparser_objects = {}
+        
+        for subparser_name in getattr(obj.__class__, '_subparsers', []):
+            try:
+                subparser_items = _get_subparser_items(obj, subparser_name)
+                
+                for subparser_key, subparser_obj in subparser_items.items():
+                    # Get all commands from this subparser
+                    for name, member in inspect.getmembers(subparser_obj):
+                        if hasattr(member, '_is_command'):
+                            if member._when and not member._when(subparser_obj):
+                                continue
+                            # Store the subparser object for this command
+                            subparser_objects[name] = subparser_obj
+            except ValueError:
+                pass
+                
+        return subparser_objects
         
     def _get_subparsers_for_obj(self, obj: Any) -> dict:
         """Get all subparsers for an object."""
@@ -196,10 +258,14 @@ class CommandWrapper:
                 # If it's a callable (but not a list/tuple), call it with the target object
                 completions[param] = completion(target_obj)
         return completions
+    
+
 
     def _find_target(self, path: str) -> tuple[Any, str, Any]:
         """Traverse the object hierarchy to find the target object and method."""
         parts = path.split('.')
+
+            
         current_obj = self._obj
         i = 0
         
@@ -246,6 +312,21 @@ class CommandWrapper:
                     i += 1
 
         method_name = parts[-1]
+
+        if self._flat and current_obj == self._obj:
+
+            if not hasattr(current_obj, method_name):
+                # method not found, check subparsers
+                for subparser_name in current_obj.__class__._subparsers:
+                    sub = getattr(current_obj, subparser_name)
+                    #check for method in subparser
+                    if hasattr(sub, method_name):
+                        method = getattr(sub, method_name)
+                        if hasattr(method, '_is_command'):
+                            return sub, method_name, method
+
+
+
         if not hasattr(current_obj, method_name):
             raise ValueError(f"Method not found: {method_name}")
             
@@ -258,6 +339,23 @@ class CommandWrapper:
     def _convert_type(self, value: str, type_hint: type, param_name: str = None, 
                      target_obj: Any = None, method: Callable = None) -> Any:
         """Convert a string value to the specified type."""
+
+        # Handle special cases for Optional types
+        if hasattr(type_hint, '__origin__'):
+            # Handle Optional types
+            if type_hint.__origin__ == Union:
+                if len(type_hint.__args__) == 2 and type_hint.__args__[1] is type(None):
+                    type_hint = type_hint.__args__[0]
+                else:
+                    raise ValueError(f"Unsupported type hint: {type_hint}")
+
+        #support multiple types 
+        validInt = False
+        hints=[]
+        strHints = str(type_hint).split('|')
+        for strHint in strHints:
+            hints.append(strHint.strip().replace('typing.','').replace('builtins.',''))
+
         # First check if we have completions for this parameter
         if param_name and method and target_obj and self._validate_options:
             completions = self._resolve_completions(method, target_obj)
@@ -270,43 +368,54 @@ class CommandWrapper:
                     except (ValueError, TypeError):
                         raise ValueError(f"Invalid value for {param_name}. Must be one of: {valid_values}")
 
-        if type_hint == bytes:
-            if value.startswith("b'") and value.endswith("'"):
-                return value[2:-1].encode('utf-8')
-            elif value.startswith('b"') and value.endswith('"'):
-                return value[2:-1].encode('utf-8')
-            else: 
-                match = regHexStr.match(value)  
-                if match:
-                    value = value.replace('0x', '').replace(' ', '')
-                    return bytes.fromhex(value)
-                return value.encode('utf-8')
-        elif type_hint == str:
-            if value.startswith('"') and value.endswith('"'):
-                return value[1:-1]
-            elif value.startswith("'") and value.endswith("'"):
-                return value[1:-1]
-            return value
-        elif type_hint == list:
+        if 'int' in hints:
+            try:
+                test = int(value)
+                validInt = True
+            except:
+                validInt = False     
 
-            if value.startswith('[') and value.endswith(']'):
-                return yaml.safe_load(value)
-            else: 
-                return yaml.safe_load(f"[{value}]")
-        elif type_hint == bool:
+                    
+
+        if 'bool' in hints:
             if value.lower() in ['true', '1', 'yes', 'on']:
                 return True
             elif value.lower() in ['false', '0', 'no', 'off']:
                 return False
             else:
                 raise ValueError(f"Invalid boolean value: {value}")
-            
+        elif 'int' in hints and validInt:
+            return int(value)
+        if 'bytes' in hints:
+            if value.startswith("b'") and value.endswith("'"):
+                return value[2:-1].encode('utf-8')
+            elif value.startswith('b"') and value.endswith('"'):
+                return value[2:-1].encode('utf-8')
+            elif value.startswith('[') and value.endswith(']'):
+                return bytes(yaml.safe_load(value))
+            else: 
+                match = regHexStr.match(value)  
+                if match:
+                    value = value.replace('0x', '').replace(' ', '')
+                    return bytes.fromhex(value)
+                return value.encode('utf-8')
+        elif 'str' in hints:
+            if value.startswith('"') and value.endswith('"'):
+                return value[1:-1]
+            elif value.startswith("'") and value.endswith("'"):
+                return value[1:-1]
+            return value
+        elif 'list' in hints:
+            if value.startswith('[') and value.endswith(']'):
+                return yaml.safe_load(value)
+            else: 
+                return yaml.safe_load(f"[{value}]")
 
         return type_hint(value)
 
     def getCompletionTree(self, base_path: str = "") -> Dict:
         """Get the complete command tree for autocompletion starting from given path."""
-        tree = {'help': {}} # Always include help command
+        tree = {} # Always include help command
         
         # Get completion info for the base path
         current_obj, commands, subparsers = self._get_completion_info_for_path(base_path)
@@ -326,36 +435,42 @@ class CommandWrapper:
                 
             # Override with actual completions if defined
             if hasattr(method, '_completions'):
-                resolved_completions = self._resolve_completions(method, current_obj)
+                # For flat mode, we need to use the correct target object
+                target_obj = self._subparser_objects.get(name, current_obj) if self._flat else current_obj
+                resolved_completions = self._resolve_completions(method, target_obj)
                 completions.update(resolved_completions)
                 
             tree[name] = completions
             
-        # Add subparsers to tree
-        if self._include_subparser_name:
-            for subparser_name, subparser_dict in subparsers.items():
-                subtree = {}
-                for key, value in subparser_dict.items():
-                    # Recursively get completion tree for subparser items
-                    new_path = f"{base_path}.{subparser_name}.{key}" if base_path else f"{subparser_name}.{key}"
-                    subtree[key] = self.getCompletionTree(new_path)
-                tree[subparser_name] = subtree
-        else:
-            for subparser_dict in subparsers.values():
-                for key, value in subparser_dict.items():
-                    # Recursively get completion tree
-                    new_path = f"{base_path}.{key}" if base_path else key
-                    tree[key] = self.getCompletionTree(new_path)
+        # Add subparsers to tree (skip in flat mode since commands are already flattened)
+        if not self._flat:
+            if self._include_subparser_name:
+                for subparser_name, subparser_dict in subparsers.items():
+                    subtree = {}
+                    for key, value in subparser_dict.items():
+                        # Recursively get completion tree for subparser items
+                        new_path = f"{base_path}.{subparser_name}.{key}" if base_path else f"{subparser_name}.{key}"
+                        subtree[key] = self.getCompletionTree(new_path)
+                    tree[subparser_name] = subtree
+            else:
+                for subparser_dict in subparsers.values():
+                    for key, value in subparser_dict.items():
+                        # Recursively get completion tree
+                        new_path = f"{base_path}.{key}" if base_path else key
+                        tree[key] = self.getCompletionTree(new_path)
                     
         return tree
     
     def getCompletions(self, partial_cmd: str, force_strings=False, use_cache=False) -> List[str]:
         """Get possible completions for a partial command."""
-
-        if partial_cmd.startswith('help '):
-            partial_cmd = partial_cmd[5:].strip()
-
-
+        
+        # Initialize state tracking
+        state = 'command'  # Can be 'command', 'arg', or 'value'
+        positional_arg = 0  # Track which positional argument we're on
+        current_tree = None
+        completions = []
+        used_keywords = []
+        
         # Get or create completion tree
         if use_cache and hasattr(self, '_tree_cache'):
             tree = self._tree_cache
@@ -363,77 +478,107 @@ class CommandWrapper:
             tree = self.getCompletionTree()
             self._tree_cache = tree
         
+        # Handle empty command
         if not partial_cmd:
-            completions = list(tree.keys())
-            if force_strings:
-                return [f'"{c}"' for c in completions]
-            return completions
+            return list(tree.keys())
             
-        # Check if we're completing parameters
-        if ' ' in partial_cmd:
-            regMatch = regCmdParams.match(partial_cmd)
-            if not regMatch:
-                return []
-                
-            cmd_path = regMatch.group(1).strip()
-            param_str = regMatch.group(2).strip()
-            
-            # Navigate through tree to find command completions
-            cmd_parts = cmd_path.split('.')
+        # Determine state and position
+        cmd_split = partial_cmd.split(' ', maxsplit=1)
+        if len(cmd_split) > 1:
+            state = 'arg'
+            arg_split = splitWithEscapes(cmd_split[1], ',')
+
+            if len(arg_split) > 0:
+                # If last argument contains :, we're completing a value
+                if ':' in arg_split[-1]:
+                    state = 'value'
+                    
+                # Count positional arguments (those without :)
+                for arg in arg_split[:-1]:  # Don't count current argument
+                    kv = splitWithEscapes(arg, ':')
+                    if len(kv) == 1:
+                        positional_arg += 1
+                    else:
+                        used_keywords.append(kv[0].strip())
+                        
+            # Navigate to command in tree
+            cmd_parts = cmd_split[0].split('.')
             current_tree = tree
             for part in cmd_parts:
                 if part not in current_tree:
                     return []
                 current_tree = current_tree[part]
                 
-            # If the command isn't found or doesn't have completions
             if not isinstance(current_tree, dict):
                 return []
                 
-            # If no param_str, return all param names
-            if not param_str:
-                completions = list(current_tree.keys())
-                if force_strings:
-                    return [f'"{c}"' for c in completions]
-                return completions
-                
-            # Split the parameters string into parts
-            parts = [p.strip() for p in param_str.split(',')]
-            
-            # If last part contains ':', we're completing values for that param
-            if ':' in parts[-1]:
-                param_name = parts[-1].split(':')[0].strip()
+            # Handle each state
+            if state == 'value':
+                # Get the parameter name before the colon
+                param_name = splitWithEscapes(arg_split[-1], ':')[0].strip()
                 if param_name in current_tree:
-                    completions = current_tree[param_name]
-                    if force_strings:
-                        return [f'"{str(c)}"' for c in completions]
-                    return completions
+                    return current_tree[param_name]
                 return []
                 
-            # Otherwise, we're completing parameter names
-            # Get list of already used parameters
-            used_params = set()
-            for part in parts[:-1]:  # Exclude current part
-                if ':' in part:
-                    used_param = part.split(':')[0].strip()
-                    used_params.add(used_param)
-                        
-            # Return unused parameters
-            completions = [p for p in current_tree.keys() if p not in used_params]
-            if force_strings:
-                return [f'"{p}"' for p in completions]
+            elif state == 'arg':
+                param_names = list(current_tree.keys())
 
-            return completions
+                for idx, param in enumerate(param_names):
+                    if param in used_keywords and idx >= positional_arg:
+                        positional_arg += 1
                 
-        # Handle path completion
+                # Get completions for current positional argument if available
+                if positional_arg < len(param_names):
+                    current_pos_param = param_names[positional_arg]
+                    if current_pos_param in current_tree:
+                        pos_completions = current_tree[current_pos_param]
+                        if isinstance(pos_completions, (list, tuple)):
+                            completions.extend(str(c) for c in pos_completions)
+                
+                # add keywords for all used positional arguments to used_keywords
+                for i in range(positional_arg):
+                    used_keywords.append(param_names[i])
+                # Add all unused parameter names
+                
+                remaining_params = [p for p in param_names if p not in used_keywords]
+                completions.extend(remaining_params)
+                
+        else:
+            # Handle command completion
+            parts = partial_cmd.split('.')
+            current_tree = tree
+            
+            # Navigate through all but the last part
+            for part in parts[:-1]:
+                if part not in current_tree:
+                    return []
+                current_tree = current_tree[part]
+                    
+            # Get completions that match the last part
+            current_part = parts[-1]
+            for name, value in current_tree.items():
+                if name.startswith(current_part):
+                    suffix = ' ' if isinstance(value, dict) and all(isinstance(v, list) for v in value.values()) else ''
+                    completions.append(name + suffix)
+        
+        if force_strings:
+            return [f'"{c}"' for c in completions]
+            
+        return completions
+
+    def getPathCompletions(self, partial_cmd: str) -> List[str]:
+        """Get possible completions for a partial command path."""
+        if not partial_cmd:
+            return list(self._commands.keys())
+            
         parts = partial_cmd.split('.')
-        current_tree = tree
+        current_tree = self._commands
         
         # Navigate through all but the last part
         for part in parts[:-1]:
             if part not in current_tree:
                 return []
-            current_tree = current_tree[part]
+            current_tree = self._get_subparsers_for_obj(current_tree[part])
             
         # Get completions that match the last part
         current_part = parts[-1]
@@ -441,14 +586,8 @@ class CommandWrapper:
         
         for name, value in current_tree.items():
             if name.startswith(current_part):
-                # Add space after leaf nodes (commands), nothing after branch nodes (subparsers)
-                suffix = ' ' if isinstance(value, dict) and all(isinstance(v, list) for v in value.values()) else ''
-                completions.append(name + suffix)
-        
-        if force_strings:
-            return [f'"{c}"' for c in completions]
-        
-
+                completions.append(name)
+                
         return completions
 
     def parseCommand(self, cmd_str: str) -> Any:
@@ -457,14 +596,27 @@ class CommandWrapper:
         cmd_parts = cmd_str.split(" ", 1)
         cmd_path = cmd_parts[0]
 
-        if cmd_path == 'help':
+        if cmd_path == 'help' and self._default_help:
+            if len(cmd_parts) == 1:
+                print(colored("Available commands:\n", color='yellow'))
+                for cmd in self._commands:
+                    helpText = getattr(self._commands[cmd], '_help', "")
+                    print(colored(f"{cmd}: {helpText}", color='yellow'))
+                return True
+
             print(colored(self.getHelp(cmd_parts[1]), color='yellow'))
             return True
         
         args_str = cmd_parts[1] if len(cmd_parts) > 1 else ""
         
-        # Find the target object and method
-        target_obj, method_name, method = self._find_target(cmd_path)
+        # Handle flat mode for simple commands
+        if self._flat and '.' not in cmd_path and cmd_path in self._subparser_objects:
+            target_obj = self._subparser_objects[cmd_path]
+            method_name = cmd_path
+            method = getattr(target_obj, method_name)
+        else:
+            # Find the target object and method using the regular path
+            target_obj, method_name, method = self._find_target(cmd_path)
         
         # Parse arguments
         pos_args, kw_args = _parse_args(args_str, self._allow_eval)
@@ -526,15 +678,22 @@ class CommandWrapper:
         """
         try:
             cmd_path = cmd_path.strip()
-            # Find the target object and method
-            target_obj, method_name, method = self._find_target(cmd_path)
+            
+            # Handle flat mode for simple commands
+            if self._flat and '.' not in cmd_path and cmd_path in self._subparser_objects:
+                target_obj = self._subparser_objects[cmd_path]
+                method_name = cmd_path
+                method = getattr(target_obj, method_name)
+            else:
+                # Find the target object and method using the regular path
+                target_obj, method_name, method = self._find_target(cmd_path)
             
             # Return the help text, defaulting to empty string if none exists
             return getattr(method, '_help', "") + "\n\n" or ""
         except ValueError as e:
             raise ValueError(f"Invalid command path: {cmd_path}") from e                            
 
-def command(*args, help: str = None, completions: Dict[str, Union[List, Callable, str]] = None):
+def command(*args, help: str = None, completions: Dict[str, Union[List, Callable, str]] = None, when: Callable = None):
     """
     Decorator to mark a method as a command.
     Can be used with or without parameters.
@@ -543,6 +702,7 @@ def command(*args, help: str = None, completions: Dict[str, Union[List, Callable
         func._is_command = True
         func._help = help or inspect.getdoc(func) or ""
         func._completions = completions or {}
+        func._when = when
         return func
 
     # Handle no-parentheses usage
@@ -562,19 +722,36 @@ def invalidatesTree(func):
     return wrapper
 
 
-def commandParser(cls=None, /, *, allow_eval: bool = False, subparsers: Optional[List[str]] = None, include_subparser_name: bool = False):
+def commandParser(cls=None, /, *, allow_eval: bool = False, subparsers: Optional[List[str]] = None, 
+                  include_subparser_name: bool = False, flat: bool = False, default_help: bool = True):
     """Class decorator to enable command parsing functionality."""
     def wrap(cls):
         # Add subparsers list as a class attribute
         cls._subparsers = subparsers or []
         cls._include_subparser_name = include_subparser_name
         cls._allow_eval = allow_eval  # Add this to store at class level
+        cls._flat = flat  # Store the flat setting
+        cls._default_help = default_help  # Store the include_help setting
         
         original_init = cls.__init__
         
         def __init__(self, *args, **kwargs):
             original_init(self, *args, **kwargs)
-            self._wrapper = CommandWrapper(self, allow_eval=allow_eval, include_subparser_name=include_subparser_name)
+            self._wrapper = CommandWrapper(
+                self, 
+                allow_eval=allow_eval, 
+                include_subparser_name=include_subparser_name,
+                flat=flat,
+                default_help=default_help
+            )
+
+        if default_help and not hasattr(cls, 'help'):
+
+            @command(completions={'cmd_path': ['$path']})
+            def help(self, cmd_path: str = "") -> str:
+                """Get help text for a command."""
+                helpStr = self._wrapper.getHelp(cmd_path)
+                print(colored(helpStr, 'yellow'))
             
         def parseCommand(self, cmd_str: str) -> Any:
             return self._wrapper.parseCommand(cmd_str)
@@ -585,6 +762,9 @@ def commandParser(cls=None, /, *, allow_eval: bool = False, subparsers: Optional
         def getCompletions(self, partial_cmd: str) -> List[str]:
             return self._wrapper.getCompletions(partial_cmd)
         
+        def getPathCompletions(self, partial_cmd: str) -> List[str]:
+            return self._wrapper.getPathCompletions(partial_cmd)
+        
         def getHelp(self, cmd_path: str) -> str:
             return self._wrapper.getHelp(cmd_path)
             
@@ -593,6 +773,7 @@ def commandParser(cls=None, /, *, allow_eval: bool = False, subparsers: Optional
         cls.parseCommand = parseCommand
         cls.getCompletionTree = getCompletionTree
         cls.getCompletions = getCompletions
+        cls.getPathCompletions = getPathCompletions
         
         return cls
         
